% LDR-LLE: LLE ALGORITHM (using K nearest neighbors) with Low-Dimensional
%                Neighborhood Representation
%
% [Y] = lle_ldr(X,K,dmax,covex_weights)
%
% X = data as D x N matrix (D = dimensionality, N = #points)
% K = number of neighbors
% d = dimension of embedding
% convex_weight = a flag for convex weights. Default 0, i.e., not convex.
%
% Y = embedding as d x N matrix
% This is a modification of the original LLE code as appears in LLE's
% homepage http://www.cs.toronto.edu/~roweis/lle/code/lle.m
% 
% The computation of the weight vectors in this version is based on first 
% projection the local neighborhood and then find the best weight vector
% with respecet to the projected neighborhood.
% (see LLE with low-dimensional neighborhood representation, 
%  http://arxiv.org/abs/0808.0780 )
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Y] = lle_ldr(X,K,d,convex_weights)

if (nargin < 4)
   convex_weights=0;
end

[D,N] = size(X);
fprintf(1,'LDR-LLE running on %d points in %d dimensions\n',N,D);


% STEP1: COMPUTE PAIRWISE DISTANCES & FIND NEIGHBORS
fprintf(1,'-->Finding %d nearest neighbours.\n',K);

X2 = sum(X.^2,1);
distance = repmat(X2,N,1)+repmat(X2',1,N)-2*X'*X;

[sorted,index] = sort(distance);
neighborhood = index(2:(1+K),:);



% STEP2: SOLVE FOR RECONSTRUCTION WEIGHTS
fprintf(1,'-->Solving for reconstruction weights.\n');

W = zeros(K,N);

%Compute the neighborhoods' weights
for ii=1:N
    % Compute the low-dimensional neighborhood
    z = X(:,neighborhood(:,ii))'-repmat(X(:,ii),1,K)'; % shift ith pt to origin
    if(K<D)
        [U,L,V]=svd(z,'econ');
    else
        [U,L,V]=svd(z);
    end
    
    if(convex_weights==0) %Compute the weights (regular)
        Ud=U(:,d+1:K);
        a=Ud'*ones(K,1);
        a=a/(a'*a);
        W(:,ii) =Ud*a;
    else %Compute the weights (convex)
        U1=U(:,1:d);L1=L(1:d,1:d);
        G=(U1*L1)*((U1*L1)');
        W(:,ii)=qpdantz(1e-2*eye(K)+(G/norm(G)),zeros(K,1),[(-1)*ones(1,K)],-1,zeros(K,1));
    end;
end;

M = sparse(1:N,1:N,ones(1,N),N,N,4*K*N);
for ii=1:N
    w = W(:,ii);
    jj = neighborhood(:,ii);
    M(ii,jj) = M(ii,jj) - w';
    M(jj,ii) = M(jj,ii) - w;
    M(jj,jj) = M(jj,jj) + w*w';
end;
options.disp = 0; options.isreal = 1; options.issym = 1;
[Y1,eigenvals] = eigs(M,d+1,0,options);
Y= Y1(:,1:d)'*sqrt(N); % bottom evect is [1,1,1,1...] with eval 0

